#define UNICODE
#define _UNICODE
#include <windows.h>
#include <crtdbg.h>
#include <string>
#include <vector>
#include "PEFile.h"
#include "LoadFile.h"


//=============================================================================


//! RXgN^BɊJǂ GetLoadBase() ŊmFB
PEFile::PEFile(const wchar_t* i_file_name)
: m_need_to_free(true),
  m_load_base(NULL),
  m_dos_header_p(NULL),
  m_nt_headers_p(NULL),
  m_section_table_p(NULL)
{
  do
  {
    // C[Wt@CɃ}bvB
    m_load_base = MapImage(i_file_name);
    if( m_load_base == NULL )
    {
      _RPTF1(_CRT_WARN, "[%S]̃[hɎs܂B", i_file_name);
      break;
    }
    _RPTF1(_CRT_WARN, "m_load_base: %p\n", m_load_base);

    // DOS wb_ւ̃|C^擾B
    m_dos_header_p = reinterpret_cast<IMAGE_DOS_HEADER*>(m_load_base);
    
    // NT wb_ւ̃|C^擾B
    m_nt_headers_p = reinterpret_cast<IMAGE_NT_HEADERS32*>(
                         m_load_base + m_dos_header_p->e_lfanew
                     );
      
    // ZNVe[uւ̃|C^擾B
    m_section_table_p = reinterpret_cast<IMAGE_SECTION_HEADER*>(
                          m_nt_headers_p + 1
                        );
  }
  while( 0 );
}


//=============================================================================


//! W[nh̃RXgN^B
PEFile::PEFile(HMODULE i_module_handle)
: m_need_to_free(false),  // VirtualFree() ̕KvȂ
  m_load_base(NULL),
  m_dos_header_p(NULL),
  m_nt_headers_p(NULL),
  m_section_table_p(NULL)
{
  do
  {
    // DOS wb_̎擾ƊmFB
    m_dos_header_p = reinterpret_cast<IMAGE_DOS_HEADER*>(i_module_handle);
    if( IsBadReadPtr(m_dos_header_p, sizeof(IMAGE_DOS_HEADER))
    ||  m_dos_header_p->e_magic != *reinterpret_cast<WORD*>("MZ") )
    {
      _RPTF0(_CRT_WARN, "LȃW[nhł͂܂B");
      break;
    }
    
    // NT wb_̎擾ƊmF
    m_nt_headers_p  = reinterpret_cast<IMAGE_NT_HEADERS32*>(
                          reinterpret_cast<BYTE*>(i_module_handle)
                        + m_dos_header_p->e_lfanew
                      );
    if( m_nt_headers_p->Signature != *reinterpret_cast<DWORD*>("PE\0\0")
    ||  m_nt_headers_p->OptionalHeader.Magic != IMAGE_NT_OPTIONAL_HDR32_MAGIC )
    {
      _RPTF0(_CRT_WARN, "PE t@Cł͂܂B\n");
      break;
    }
    
    // ZNVe[uւ̃|C^擾B
    m_section_table_p = reinterpret_cast<IMAGE_SECTION_HEADER*>(m_nt_headers_p + 1);
    // m_load_base  i_module_handle ĂB
    m_load_base = reinterpret_cast<BYTE*>(i_module_handle);
  }
  while( 0 );
}


//=============================================================================


//! fXgN^
PEFile::~PEFile()
{
  if( m_need_to_free && m_load_base )
  {
    VirtualFree(m_load_base, 0, MEM_RELEASE);
    m_load_base = NULL;
  }
}


//=============================================================================


//! [hAhX擾B
BYTE* PEFile::GetLoadBase()
{
  return m_load_base;
}


//=============================================================================


//! DOS wb_ւ̃|C^擾B
IMAGE_DOS_HEADER* PEFile::GetDosHeaderP()
{
  return GetLoadBase() ? m_dos_header_p : NULL;
}


//=============================================================================


//! NT wb_ւ̃|C^擾B
IMAGE_NT_HEADERS32* PEFile::GetNtHeadersP()
{
  return GetLoadBase() ? m_nt_headers_p : NULL;
}


//=============================================================================


//! t@Cwb_ւ̃|C^擾B
IMAGE_FILE_HEADER* PEFile::GetFileHeaderP()
{
  return GetLoadBase() ? &m_nt_headers_p->FileHeader : NULL;
}


//=============================================================================


//! IvViwb_ւ̃|C^擾B
IMAGE_OPTIONAL_HEADER32* PEFile::GetOptionalHeaderP()
{
  return GetLoadBase() ? &m_nt_headers_p->OptionalHeader : NULL;
}


//=============================================================================


//! i_index Ԗڂ̃ZNVwb_ւ̃|C^擾B
IMAGE_SECTION_HEADER* PEFile::GetSectionHeader(int i_index)
{
  // i_index ZNVɎ܂Ă邩mFB
  if( GetLoadBase() && 0 <= i_index && i_index < GetFileHeaderP()->NumberOfSections )
    return &m_section_table_p[i_index];
  else
    return NULL;
}


//=============================================================================


//! i_index Ԗڂ̃f[^fBNgGgwf[^ւ̃|C^擾B
BYTE* PEFile::GetDirEntryDataP(int i_index)
{
  if( GetLoadBase()
  &&  0 <= i_index && i_index < IMAGE_NUMBEROF_DIRECTORY_ENTRIES
  &&  GetOptionalHeaderP()->DataDirectory[i_index].VirtualAddress
  &&  GetOptionalHeaderP()->DataDirectory[i_index].Size )
  {
    return m_load_base + GetOptionalHeaderP()->DataDirectory[i_index].VirtualAddress;
  }
  else
  {
    return NULL;
  }
}


//=============================================================================


//! i_index Ԗڂ̃f[^fBNgGgwf[^ւ̃TCY擾B
DWORD PEFile::GetDirEntryDataSize(int i_index)
{
  if( GetLoadBase() && 0 <= i_index && i_index < IMAGE_NUMBEROF_DIRECTORY_ENTRIES )
    return GetOptionalHeaderP()->DataDirectory[i_index].Size;
  else
    return 0;
}


//=============================================================================


//! C|[gfBNgւ̃|C^擾B
IMAGE_IMPORT_DESCRIPTOR* PEFile::GetImportDirP()
{
  return reinterpret_cast<IMAGE_IMPORT_DESCRIPTOR*>(
           GetDirEntryDataP(IMAGE_DIRECTORY_ENTRY_IMPORT)
         );
}


//=============================================================================


//! C|[gĂ DLL 擾B
std::vector<const char*> PEFile::GetImportDllNames()
{
  std::vector<const char*>  retval;
  IMAGE_IMPORT_DESCRIPTOR*  imp_descs_p = GetImportDirP();
  
  if( imp_descs_p )
  {
    for(int i = 0; imp_descs_p[i].FirstThunk != 0; i++)
    {
      retval.push_back(GetDataP<const char*>(imp_descs_p[i].Name));
    }
  }
  
  return retval;
}


//=============================================================================


//! w肵 DLL C|[gĂV{擾B
std::vector<ImportSymbol> PEFile::GetImportSymbols(const char* i_dll_name)
{
  std::vector<ImportSymbol>   retval;
  IMAGE_IMPORT_DESCRIPTOR*    imp_descs_p = GetImportDirP();

  // C|[gl[e[uĂ邩APEFile NXOŃ[h
  // C[Wt@Ĉݗ񋓉\B
  if( imp_descs_p && (imp_descs_p[0].OriginalFirstThunk || m_need_to_free) )
  {
    for(int i = 0; imp_descs_p[i].FirstThunk != 0; i++)
    {
      const char* dll_name = GetDataP<const char*>(imp_descs_p[i].Name);
      if( stricmp(dll_name, i_dll_name) == 0 )  // w肳ꂽ DLL 𔭌!
      {
        // C|[gAhXe[uւ̃|C^擾B
        DWORD* iat_p = GetDataP<DWORD*>(imp_descs_p[i].FirstThunk);
        // C|[gl[e[uւ̃|C^擾B
        // INT ݂Ȃꍇ͑ IAT ւ̃|C^gpB
        DWORD* int_p = imp_descs_p[i].OriginalFirstThunk
                     ? GetDataP<DWORD*>(imp_descs_p[i].OriginalFirstThunk)
                     :  iat_p;

        // C|[gV{
        for(int table_index = 0; int_p[table_index] != 0; table_index++)
        {
          // INT ̃R[hC[WɎ܂Ă邩mFB
          if( int_p[table_index] < GetOptionalHeaderP()->SizeOfImage )
          {
            ImportSymbol  symbol = { 0 };

            symbol.iat_entry_rva = imp_descs_p[i].FirstThunk
                                 + sizeof(DWORD) * table_index;
            symbol.iat_entry_p   = reinterpret_cast<void**>(&iat_p[table_index]);

            // ɂC|[g
            if( IMAGE_SNAP_BY_ORDINAL32(int_p[table_index]) ) 
            {
              symbol.ordinal = static_cast<WORD>(IMAGE_ORDINAL32(int_p[table_index]));
            }
            // OɂC|[g
            else
            {
              IMAGE_IMPORT_BY_NAME*
                by_name_p = GetDataP<IMAGE_IMPORT_BY_NAME*>(int_p[table_index]);
              symbol.hint = by_name_p->Hint;
              symbol.name = reinterpret_cast<char*>(by_name_p->Name);
            }

            retval.push_back(symbol);
          }
        }
      }
    }
  }
  
  return retval;
}


//=============================================================================


//! C|[gV{tbNBꍇ IAT Ggւ̃|C^ԂB
void** PEFile::HookImportSymbol(void* i_old_p, void* i_new_p)
{
  void**                    retval      = NULL;
  IMAGE_IMPORT_DESCRIPTOR*  imp_descs_p = GetImportDirP();
  
  if( imp_descs_p )
  {
    for(int i = 0; imp_descs_p[i].FirstThunk != 0; i++)
    {
      // C|[gAhXe[uւ̃|C^擾B
      void**  iat_p = GetDataP<void**>(imp_descs_p[i].FirstThunk);

      for(int j = 0; iat_p[j] != 0; j++)
      {
        if( iat_p[j] == reinterpret_cast<FARPROC>(i_old_p) )
        {
          DWORD old_protect = 0;
          VirtualProtect(&iat_p[j], sizeof(FARPROC), PAGE_READWRITE, &old_protect);

          retval   = &iat_p[j];
          iat_p[j] = reinterpret_cast<FARPROC>(i_new_p);

          VirtualProtect(&iat_p[j], sizeof(FARPROC), old_protect, &old_protect);
          // break; // break ȂBSĂ IAT 𑖍B
        }
      }

      break;
    }
  }
  
  return retval;
}


//=============================================================================


//! GNX|[gfBNgւ̃|C^擾B
IMAGE_EXPORT_DIRECTORY* PEFile::GetExportDirP()
{
  return reinterpret_cast<IMAGE_EXPORT_DIRECTORY*>(
           GetDirEntryDataP(IMAGE_DIRECTORY_ENTRY_EXPORT)
         );
}


//=============================================================================


//! i_index Ԗڂ̃GNX|[g̖O擾Bsꍇ "" ԂB
const char* PEFile::FindExportSymbolName(size_t i_index)
{
  const char*               retval        = "";
  IMAGE_EXPORT_DIRECTORY*   export_dir_p  = GetExportDirP();
  
  if( export_dir_p )
  {
    DWORD* name_rvas_p = GetDataP<DWORD*>(export_dir_p->AddressOfNames);
    WORD*  indices_p   = GetDataP<WORD*>(export_dir_p->AddressOfNameOrdinals);
    for(DWORD i = 0; i < export_dir_p->NumberOfNames; i++)
    {
      if( indices_p[i] == i_index )
      {
        retval = GetDataP<const char*>(name_rvas_p[i]);
        break;
      }
    }
  }
  
  return retval;
}


//=============================================================================


//! i_index Ԗڂ̃GNX|[gV{ RVA 擾B
DWORD PEFile::GetExportSymbolRVA(size_t i_index)
{
  DWORD                   retval        = 0;
  IMAGE_EXPORT_DIRECTORY* export_dir_p  = GetExportDirP();
  
  if( export_dir_p
  &&  0 <= i_index && i_index < export_dir_p->NumberOfFunctions )
  {
    retval = GetDataP<DWORD*>(export_dir_p->AddressOfFunctions)[i_index];
  }
  
  return retval;
}


//=============================================================================


//! GNX|[g]Ă邩ǂׂB
bool PEFile::IsExportForwarded(DWORD i_symbol_rva)
{
  bool  retval = false;

  if( GetLoadBase() )
  {
    IMAGE_DATA_DIRECTORY& dir_entry = GetOptionalHeaderP()->DataDirectory[
                                        IMAGE_DIRECTORY_ENTRY_EXPORT
                                      ];

    // GNX|[gfBNg݂Ă RVA ͈͓̔B
    retval = dir_entry.VirtualAddress && dir_entry.Size
          && dir_entry.VirtualAddress <= i_symbol_rva
          && i_symbol_rva < dir_entry.VirtualAddress + dir_entry.Size;
  }
  
  return retval;
}


//=============================================================================


//! GNX|[gV{擾B
std::vector<ExportSymbol> PEFile::GetExportSymbols()
{
  std::vector<ExportSymbol> retval;
  IMAGE_EXPORT_DIRECTORY*   export_dir_p = GetExportDirP();
  
  if( export_dir_p )
  {
    for(size_t i = 0; i < export_dir_p->NumberOfFunctions; i++)
    {
      ExportSymbol  symbol = { 0 };

      symbol.symbol_rva   = GetExportSymbolRVA(i);
      if( symbol.symbol_rva )
      {
        symbol.symbol_p     = GetDataP<void*>(symbol.symbol_rva);
        symbol.ordinal      = i + export_dir_p->Base;
        symbol.name         = FindExportSymbolName(i);
        symbol.forwarded_to = IsExportForwarded(symbol.symbol_rva)
                            ? reinterpret_cast<char*>(symbol.symbol_p)
                            : "";

        retval.push_back(symbol);
      }
    }
  }
  while( 0 );
  
  return retval;
}


//=============================================================================


//! GNX|[gV{ւ̃|C^擾B
void* PEFile::GetExportSymbolPFromOrdinal(WORD i_ordinal)
{
  void*                     retval        = NULL;
  IMAGE_EXPORT_DIRECTORY*   export_dir_p  = GetExportDirP();
  
  if( export_dir_p )
  {
    DWORD symbol_rva = GetExportSymbolRVA(i_ordinal - export_dir_p->Base);
    if( symbol_rva )
    {
      if( IsExportForwarded(symbol_rva) )  // ]ĂꍇB
      {
        char* forwarded_to       = GetDataP<char*>(symbol_rva);
        char* symbol_name_p      = strstr(forwarded_to, ".") + 1;
        char  dll_name[MAX_PATH] = { 0 };

        // DLL Rs[Ċgqt^B
        memcpy(dll_name, forwarded_to, symbol_name_p - forwarded_to);
        strncat(dll_name, "DLL", sizeof(dll_name) - 1);
        
        HMODULE module_handle = GetModuleHandleA(dll_name);
        if( module_handle == NULL )
          module_handle = LoadLibraryA(dll_name);
        
        if( module_handle )
          retval = PEFile(module_handle).GetExportSymbolPFromName(symbol_name_p);
      }
      else  // ]ĂȂꍇ
      {
        retval = GetDataP<void*>(symbol_rva);
      }
    }
  }
  
  return retval;
}


//=============================================================================


//! GNX|[gV{ւ̃|C^擾B
void* PEFile::GetExportSymbolPFromName(const char* i_name)
{
  void*                     retval        = NULL;
  IMAGE_EXPORT_DIRECTORY*   export_dir_p  = GetExportDirP();
  
  if( export_dir_p )
  {
    DWORD*  name_rvas_p = GetDataP<DWORD*>(export_dir_p->AddressOfNames);
    WORD*   indices_p   = GetDataP<WORD*>(export_dir_p->AddressOfNameOrdinals);
    
    // oCiT[`sB
    int     left    = 0;
    int     right   = export_dir_p->NumberOfNames - 1;
    while( left <= right )
    {
      int   middle      = (left + right) / 2;
      char* name        = GetDataP<char*>(name_rvas_p[middle]);
      int   comp_result = strcmp(i_name, name);

      if( comp_result == 0 )  // !
      {
        // CfbNXɕϊāA|C^擾B
        retval = GetExportSymbolPFromOrdinal(
                   static_cast<WORD>(indices_p[middle] + export_dir_p->Base)
                 );
        break;
      }
      else if( comp_result > 0 )
      {
        left = middle + 1;
      }
      else
      {
        right = middle - 1;
      }
    }
  }
  
  return retval;
}


//=============================================================================


//! IMAGE_BASE_RELOCATION ̑Ɏgp\́Boƃ\bhǉB
struct BaseRelocation
{
  DWORD   virtual_address;  //!< ĔzůJn RVA
  DWORD   size_of_block;    //!< ̍ĔzuubÑTCY
  WORD    fixup_records[1]; //!< 4rbgŃ^CvA12rbgŃItZbg
  
  //! fixup_records ̃R[h擾B
  size_t GetFixupLength()
  {
    return (size_of_block - sizeof(IMAGE_BASE_RELOCATION)) / sizeof(WORD);
  }
  
  //! i_index Ԗڂ fixup_records ̍Ĕzu^Cv擾B
  DWORD GetType(int i_index)
  {
    return fixup_records[i_index] >> 12;
  }
  
  //! i_index Ԗڂ RVA 擾B
  DWORD GetRVA(int i_index)
  {
    return (fixup_records[i_index] & ((1 << 12) - 1)) + virtual_address;
  }
};


//=============================================================================


//! x[XĔzu擾B
std::vector<DWORD*> PEFile::GetBaseRelocations()
{
  std::vector<DWORD*> retval;
  DWORD offset        = 0;
  DWORD reloc_size    = GetDirEntryDataSize(IMAGE_DIRECTORY_ENTRY_BASERELOC);
  BYTE* first_reloc_p = GetDirEntryDataP(IMAGE_DIRECTORY_ENTRY_BASERELOC);
  
  if( first_reloc_p )
  {
    while( offset < reloc_size )
    {
      BaseRelocation&
        reloc = *reinterpret_cast<BaseRelocation*>(first_reloc_p + offset);
      
      for(size_t i = 0; i < reloc.GetFixupLength(); i++)
      {
        if( reloc.GetType(i) == IMAGE_REL_BASED_HIGHLOW )
          retval.push_back(GetDataP<DWORD*>(reloc.GetRVA(i)));
      }
      
      offset += reloc.size_of_block;
    }
  }
  
  return retval;
}


//=============================================================================


//! C|[gV{̃oCh
bool PEFile::BindImportSymbols()
{
  bool                      retval    = true;
  std::vector<const char*>  dll_names = GetImportDllNames();
  
  for(size_t dll_index = 0; retval && dll_index < dll_names.size(); dll_index++)
  {
    HMODULE dll_handle = GetModuleHandleA(dll_names[dll_index]);
    // GetModuleHandle() sȂ LoadLibrary() Ă݂B
    if( dll_handle == NULL )
      dll_handle = LoadLibraryA(dll_names[dll_index]);

    // nh擾łȂȂ玸sB
    if( dll_handle == NULL )
    {
      retval = false;
      break;
    }

    std::vector<ImportSymbol> symbols = GetImportSymbols(dll_names[dll_index]);
    for(size_t symbol_index = 0; symbol_index < symbols.size(); symbol_index++)
    {
      *symbols[symbol_index].iat_entry_p = GetProcAddress(
                                             dll_handle,
                                             symbols[symbol_index].name
                                           );
      // V{łȂB
      if( *symbols[symbol_index].iat_entry_p == NULL )
      {
        retval = false;
        break;
      }
    }
  }
  
  return retval;
}


//=============================================================================


//! [[hsBGg|Cgւ̃|C^ԂB
void* PEFile::PseudoLoad(void* i_rebase_address)
{
  void*                 retval    = NULL;
  std::vector<DWORD*>   relocs    = GetBaseRelocations();

  if( relocs.size() && BindImportSymbols() )
  {
    DWORD rebase_address = i_rebase_address
                         ? reinterpret_cast<DWORD>(i_rebase_address)
                         : reinterpret_cast<DWORD>(m_load_base);

    // x[XĔzusB
    DWORD diff = rebase_address - GetOptionalHeaderP()->ImageBase;
    for(size_t i = 0; i < relocs.size(); i++)
    {
      *relocs[i] += diff;
    }
    
    retval = reinterpret_cast<BYTE*>(
               rebase_address + GetOptionalHeaderP()->AddressOfEntryPoint
             );
  }
  
  return retval;
}


//=============================================================================


//! C[Wt@CɃ}bsOB
BYTE* PEFile::MapImage(const wchar_t* i_file_name)
{
  BYTE*     retval          = NULL;
  
  do
  {
    // t@CŜǂݍށB
    std::vector<BYTE> file_data;
    if( LoadFile(i_file_name, file_data) == false )
      break;

    // DOS wb_̎擾ƊmFB
    IMAGE_DOS_HEADER* dos_header_p = reinterpret_cast<IMAGE_DOS_HEADER*>(
                                       &file_data[0]
                                     );
    if( dos_header_p->e_magic != *reinterpret_cast<WORD*>("MZ")
    ||  dos_header_p->e_lfanew == 0 )
    {
      _RPTF0(_CRT_WARN, "s\t@Cł͂܂B\n");
      break;
    }
    
    // NT wb_̎擾ƊmFB
    IMAGE_NT_HEADERS32* nt_headers_p = reinterpret_cast<IMAGE_NT_HEADERS*>(
                                         &file_data[dos_header_p->e_lfanew]
                                       );
    if( nt_headers_p->Signature != *reinterpret_cast<DWORD*>("PE\0\0")
    ||  nt_headers_p->OptionalHeader.Magic != IMAGE_NT_OPTIONAL_HDR32_MAGIC )
    {
      _RPTF0(_CRT_WARN, "PE t@Cł͂܂B\n");
      break;
    }
    
    // ZNVe[uւ̃|C^擾B
    IMAGE_SECTION_HEADER* section_table_p = reinterpret_cast<IMAGE_SECTION_HEADER*>(
                                              nt_headers_p + 1
                                            );
    
    // SizeOfImage mہB
    retval  = reinterpret_cast<BYTE*>(
                VirtualAlloc(
                  NULL,
                  nt_headers_p->OptionalHeader.SizeOfImage,
                  MEM_COMMIT,
                  PAGE_READWRITE
                )
              );
    if( retval == NULL )
    {
      _RPTF1(_CRT_WARN, "̊蓖[%dbytes]Ɏs܂B\n",
             nt_headers_p->OptionalHeader.SizeOfImage
             );
      break;
    }

    // wb_Rs[
    memcpy(retval, &file_data[0], nt_headers_p->OptionalHeader.SizeOfHeaders);
    // ZNVf[^Rs[
    for(DWORD i = 0; i < nt_headers_p->FileHeader.NumberOfSections; i++)
    {
      // ZNVf[^ꍇB
      if( section_table_p[i].PointerToRawData )
      {
        memcpy(&retval[section_table_p[i].VirtualAddress],
               &file_data[section_table_p[i].PointerToRawData],
               section_table_p[i].SizeOfRawData
              );
      }
    }
  }
  while( 0 );
  
  return retval;
}


//=============================================================================


